class TextXLSError(Exception):
    """Indicates a generic textX-LS-core error."""

    pass


class GenerateExtensionError(TextXLSError):
    """Indicates an error while generating an extension."""

    def __init__(self, target, cmd_args):
        super().__init__(
            "Failed to generate the extension for '{}' with following arguments: '{}'.".format(
                target, cmd_args
            )
        )
        self.target = target
        self.cmd_args = cmd_args


class GenerateSyntaxHighlightError(TextXLSError):
    """Indicates an error while generating a syntax highlighting data."""

    def __init__(self, project_name, target):
        super().__init__(
            "Failed to generate syntax highlighting for project '{}' and target '{}'.".format(
                project_name, target
            )
        )
        self.project_name = project_name
        self.target = target


class InstallTextXProjectError(TextXLSError):
    """Indicates an error while installing a textX project."""

    def __init__(self, project_name, dist_location, detailed_err_msg):
        super().__init__(
            "Failed to install project: {} ({}).".format(project_name, dist_location)
        )
        self.project_name = project_name
        self.dist_location = dist_location
        self.detailed_err_msg = detailed_err_msg


class UninstallTextXProjectError(TextXLSError):
    """Indicates an error while uninstalling a textX project."""

    def __init__(self, project_name, detailed_err_msg):
        super().__init__("Failed to uninstall project: {}".format(project_name))
        self.project_name = project_name
        self.detailed_err_msg = detailed_err_msg


class LanguageNotRegistered(TextXLSError):
    """Indicates an error if language can't be parsed with any of registered metamodels.
    """

    def __init__(self, file_name):
        super().__init__(
            "There are no metamodels that can parse '{}'.".format(file_name)
        )


class MultipleLanguagesError(TextXLSError):
    """Indicates an error if language can be parsed with multiple registered metamodels.
    """

    def __init__(self, file_name):
        super().__init__(
            "Multiple languages can parse '{}'. Install appropriate extension and select language id.".format(
                file_name
            )
        )



from typing import Optional

from textx import LanguageDesc


class TextXLanguage:
    """Represents a textX language without metamodel attribute.

    NOTE: Attributes are in camelCase notation needed for client-server automatic
          (de)serialization

    Attributes:
        name (str): language name
        description (str): language description or language name
        projectName (str): project that registers this language

    """

    def __init__(self, lang_desc: LanguageDesc):
        self.name = lang_desc.name
        self.description = lang_desc.description or self.name
        self.projectName = lang_desc.project_name


class TextXProject:
    """Represents a textX project.

    NOTE: Attributes are in camelCase notation needed for client-server automatic
          (de)serialization

    Attributes:
        projectName (str): project name
        editable (bool): if project is installed in editable mode
        distLocation (str): project distribution location
        version (str): project version
        languages (list): languages that are registered by this project

    """

    def __init__(
        self,
        project_name: str,
        editable: bool,
        version: str,
        dist_location: Optional[str] = None,
    ):
        self.projectName = project_name
        self.editable = editable
        self.distLocation = dist_location
        self.version = version
        self.languages = []

    def add_language(self, language: TextXLanguage):
        self.languages.append(language)



import asyncio
from os import linesep
from os.path import basename, isdir, isfile, join, splitext
from typing import Callable, Iterable, Optional, Tuple


def compare_project_names(p1: str, p2: str) -> bool:
    """Compares project names replacing `-` with `_`.

    Args:
        p1: project name
        p2: project name to compare to
    Returns:
        True if project names are the same, otherwise False
    Raises:
        None

    """

    def replace(name):
        return name.replace("-", "_").lower()

    return replace(p1) == replace(p2)


def dist_is_editable(dist_location: str, project_name: str) -> bool:
    """Checks if distribution is installed in editable mode.

    Args:
        dist_location: location of python package
        project_name: project name
    Returns:
        True if project is installed in editable mode (with `-e` flag), otherwise False
    Raises:
        None

    """

    def _check_egg(name):
        return isdir(join(dist_location, name + ".egg-info")) or isfile(
            join(dist_location, name + ".egg-link")
        )

    return _check_egg(project_name) or _check_egg(project_name.replace("-", "_"))


def get_file_extension(uri: str) -> str:
    """Returns file extension (without dot) or file name.

    Args:
        uri: file uri
    Returns:
        File extension
    Raises:
        None

    """
    try:
        ext = splitext(uri)[1]
        return ext[1:] if ext else basename(uri)
    except (AttributeError, IndexError, TypeError):
        return ""


def get_project_name_and_version(folder_or_wheel: str) -> Tuple[str, str]:
    """Returns project name for a given project folder (with setup.py)
    or wheel file.

    Args:
        folder_or_wheel: path to the folder or wheel of textx language project
    Returns:
        Project name and version
    Raises:
        None

    """
    project_name = None
    version = None
    # Folder with setup.py inside (project)
    setuppy = join(folder_or_wheel, "setup.py")
    if isfile(setuppy):
        from unittest import mock
        import setuptools

        with mock.patch.object(setuptools, "setup") as mock_setup:
            with open(setuppy, "r") as f:
                exec(f.read())
            _, kwargs = mock_setup.call_args
        project_name, version = kwargs.get("name", None), kwargs.get("version", None)
    elif isfile(folder_or_wheel) and folder_or_wheel.endswith(".whl"):  # wheel file
        from wheel_inspect import inspect_wheel

        project_info = inspect_wheel(folder_or_wheel)
        project_name, version = (
            project_info.get("project", None),
            project_info.get("version", None),
        )

    return project_name, version


async def run_async(
    cmd: Iterable, msg_handler: Optional[Callable] = None, cwd: Optional[str] = None
) -> int:
    """Runs process using `asyncio.create_subprocess_exec`.

    Args:
        cmd: command to execute
        msg_handler: a callable which is called with message argument when process writes to stdout
        cwd: current working directory
    Returns:
        Return code and output of a process
    Raises:
        None

    """
    process = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.STDOUT, cwd=cwd
    )

    output = []

    async def _listener(process, msg_handler):
        while not process.stdout.at_eof():
            line = await process.stdout.readline()
            if line is not None:
                line = line.decode().rstrip()
                output.append(line)
                # Call message handler
                if msg_handler:
                    msg_handler(line)
        return process

    process = await asyncio.ensure_future(_listener(process, msg_handler))

    return process.returncode, linesep.join(output)



import sys
from typing import Callable, List, Optional, Tuple

from pkg_resources import DistributionNotFound, get_distribution
from textx import (
    LanguageDesc,
    clear_language_registrations,
    language_description,
    language_descriptions,
    languages_for_file,
)
from textx.exceptions import TextXRegistrationError

from ..exceptions import (
    InstallTextXProjectError,
    LanguageNotRegistered,
    MultipleLanguagesError,
    UninstallTextXProjectError,
)
from ..models import TextXLanguage, TextXProject
from ..utils import (
    compare_project_names,
    dist_is_editable,
    get_project_name_and_version,
    run_async,
)

BUILT_IN_LANGUAGES = ["textx", "txcl"]


def get_languages() -> List[TextXLanguage]:
    """Returns all registered languages.

    Args:
        None
    Returns:
        A list of registered textX languages
    Raises:
        None

    """
    return [
        TextXLanguage(lang)
        for lang in language_descriptions().values()
        if lang.name.lower() not in BUILT_IN_LANGUAGES
    ]


def get_languages_by_project_name(project_name: str) -> List[TextXLanguage]:
    """Returns all project languages.

    Args:
        project_name: textX project
    Returns:
        A list of textX project languages
    Raises:
        None

    """
    return [
        lang
        for lang in get_languages()
        if compare_project_names(lang.projectName, project_name)
    ]


def get_language_desc(
    language_name: str, file_name: Optional[str] = None
) -> LanguageDesc:
    """Returns a language info by a given language name or file extension.

    Args:
        language_name: textX language name (id)
        file_name: file path of active document
    Returns:
        A language info (`LanguageDesc`)
    Raises:
        MultipleLanguagesError: If there are multiple registered metamodels that can
                                parse a language
        LanguageNotRegistered: If there are no registered metamodels that can parse a
                               language

    """
    lang_desc = None
    try:
        lang_desc = language_description(language_name)
    except TextXRegistrationError:
        # When client does not set language id/name, i.e. extension is not
        # installed, then try to load by file extension
        lang_descs = languages_for_file(file_name)
        lang_descs_len = len(lang_descs)
        if lang_descs_len > 1:
            raise MultipleLanguagesError(file_name)

        lang_desc = lang_descs.pop() if lang_descs_len == 1 else None

    if lang_desc is None:
        raise LanguageNotRegistered(file_name)

    return lang_desc


def get_projects(with_langs: Optional[bool] = True) -> List[TextXProject]:
    """Returns all textX language projects.

    NOTE: Project is a python package which registers textX languages and generators
    via entry points

    Args:
        with_langs: include language instances
    Returns:
        A list of textX projects
    Raises:
        None

    """
    projects = {}
    for lang in get_languages():
        project_name = lang.projectName
        if project_name not in projects:
            dist = get_distribution(project_name)
            dist_location = dist.location
            editable = dist_is_editable(dist_location, project_name)
            projects[project_name] = TextXProject(
                project_name, editable, dist.version, dist_location
            )
        if with_langs is True:
            projects[project_name].add_language(lang)
    return projects


async def install_project_async(
    folder_or_wheel: str,
    python_path: str,
    editable: Optional[str] = False,
    msg_handler: Optional[Callable] = None,
) -> Tuple[str, str, str]:
    """Installs textX project.

    Args:
        folder_or_wheel: path to the folder or wheel of textX language project
        python_path: python path from virtual environment that extension is using
        editable: flag if project should be installed in editable mode
        msg_handler: a callable which is called with message argument when process writes to stdout
    Returns:
        A tuple of project name, version and package dist location if project is installed successfully
    Raises:
        InstallTextXProjectError: If project is not installed

    """
    project_name, version = get_project_name_and_version(folder_or_wheel)
    dist_location = None

    cmd = [python_path, "-m", "pip", "install", folder_or_wheel]
    if editable:
        cmd.insert(4, "-e")

    retcode, output = await run_async(cmd, msg_handler)

    # Not installed
    if retcode != 0:
        raise InstallTextXProjectError(project_name, dist_location, output)

    # Manually add package to sys.path if installed with -e flag
    if editable and folder_or_wheel not in sys.path:
        sys.path.append(folder_or_wheel)

    clear_language_registrations()

    # Checks if language with the same name already exist
    try:
        language_descriptions()
        dist_location = get_distribution(project_name).location
    except Exception as e:
        await uninstall_project_async(project_name, python_path)
        raise InstallTextXProjectError(project_name, dist_location, output) from e

    return project_name, version, dist_location


async def uninstall_project_async(
    project_name: str, python_path: str, msg_handler: Optional[Callable] = None
) -> None:
    """Uninstalls textX project.

    Args:
        project_name: project name
        python_path: python path from virtual environment that extension is using
        msg_handler: a callable which is called with message argument when process writes to stdout
    Returns:
        None
    Raises:
        UninstallTextXProjectError: If project is not uninstalled
        DistributionNotFound: If package dist is not found

    """
    cmd = [python_path, "-m", "pip", "uninstall", project_name, "-y"]

    # Get dist location before uninstalling with pip
    try:
        dist_location = get_distribution(project_name).location
    except DistributionNotFound as e:
        raise UninstallTextXProjectError(project_name, "") from e

    # Call pip uninstall command
    retcode, output = await run_async(cmd, msg_handler)
    if retcode != 0:
        raise UninstallTextXProjectError(project_name, output)

    # Manually remove package from sys.path if needed
    if dist_location in sys.path:
        sys.path.remove(dist_location)

    clear_language_registrations()



from typing import List, Union

from textx.exceptions import TextXError
from textx.metamodel import TextXMetaMetaModel, TextXMetaModel


def validate(
    metamodel: Union[TextXMetaMetaModel, TextXMetaModel], model: str, file_path: str
) -> List[TextXError]:
    """Validates given model.

    NOTE: For now returned list will contain maximum one error, since textX does not
          have built-in error recovery mechanism.

    Args:
        metamodel: language metamodel
        model: model
        file_path: A path to the `model` file
    Returns:
        A list of textX errors or empty list if model is valid
    Raises:
        None

    """
    errors = []
    try:
        metamodel.model_from_str(model, file_name=file_path)
    except TextXError as e:
        errors.append(e)
    return errors



# flake8: noqa
import codecs
import os
from platform import python_version

from setuptools import find_packages, setup

PACKAGE_NAME = "textx-ls-core"
VERSION = "0.1.0"
AUTHOR = "Daniel Elero"
AUTHOR_EMAIL = "danixeee@gmail.com"
DESCRIPTION = (
    "a core language server logic for domain specific languages based on textX"
)
KEYWORDS = "textX DSL python domain specific languages"
LICENSE = "MIT"
URL = "https://github.com/textX/textX-LS/core"

packages = find_packages()

print("packages:", packages)

README = codecs.open(
    os.path.join(os.path.dirname(__file__), "README.md"), "r", encoding="utf-8"
).read()

dev_require = ["bandit==1.5.1", "flake8==3.7.7", "textx_gen_vscode>=0.1.2"]

tests_require = ["coverage==4.5.3", "pytest==4.3.1", "pytest-cov==2.6.1"]

# pip install textx_ls_core[vscode]
vscode_require = ["textx_gen_vscode>=0.1.2"]


if python_version().startswith("3.6"):  # For python 3.6
    dev_require.append("black")


setup(
    name=PACKAGE_NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    keywords=KEYWORDS,
    license=LICENSE,
    packages=packages,
    include_package_data=True,
    package_data={"": ["*.tx"]},
    install_requires=["textX>=2.1.0", "click==7.0", "wheel_inspect==1.3.0"],
    extras_require={
        "dev": dev_require,
        "test": tests_require,
        "vscode": vscode_require,
    },
    tests_require=tests_require,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Information Technology",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



from typing import Any, List, Mapping, Optional

from textx import GeneratorDesc, generator_descriptions, generator_for_language_target

from ..exceptions import GenerateExtensionError, GenerateSyntaxHighlightError
from .projects import get_language_desc, get_languages_by_project_name


def generate_extension(target: str, dest_dir: str, **cmd_args: Optional[dict]) -> None:
    """Generates client installable extension.

    Args:
        target: target (client) to generate extension for (e.g. `vscode`)
        dest_dir: destination directory where extension will be written
        cmd_args: other optional arguments
    Returns:
        None
    Raises:
        GenerateExtensionError: if generator does not exists (`TextXRegistrationError`)
                                or any other error that happened while generating
                                the extension

    """
    try:
        extension_gen = generator_for_language_target("textX", target)
        extension_gen(None, None, dest_dir, **cmd_args)
    except Exception as e:
        raise GenerateExtensionError(target, cmd_args) from e


def generate_syntaxes(
    project_name: str, target: str, **cmd_args: Optional[dict]
) -> Mapping[str, Any]:
    """Generates syntax highlighting files for all project languages.

    NOTE: This is called on grammar changes and it collects new language keywords.
    (project must be installed in editable mode)

    Args:
        project_name: project for which to generate syntax highlightings
        target: syntax highlight file type (e.g. `vscode`)
        cmd_args: other optional arguments
    Returns:
        A map where keys are language names and values are corresponding syntax
        highlighting files
    Raises:
        GenerateSyntaxHighlightError: if generator does not exists (`TextXRegistrationError`)
                                      or any other error that happened while generating
                                      syntax highlighting files

    """
    try:
        syntax_gen = generator_for_language_target("textX", target)

        lang_syntax_map = {}
        for lang in get_languages_by_project_name(project_name):
            lang_name = lang.name.lower()
            lang_desc = get_language_desc(lang_name)
            syntax_file = syntax_gen(
                None, lang_desc.metamodel(), **{"name": lang_name, **cmd_args}
            )
            if syntax_file:
                lang_syntax_map[lang_name] = syntax_file
        return lang_syntax_map
    except Exception as e:
        raise GenerateSyntaxHighlightError(project_name, target) from e


def get_generators() -> List[GeneratorDesc]:
    """Returns all registered generators.

    Args:
        None
    Returns:
        A list of registered textX generators
    Raises:
        None

    """
    return [
        gen
        for target_gens in generator_descriptions().values()
        for gen in target_gens.values()
    ]
